/**
 * Copyright (C) 2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.perseus.concurrency.lib;

import java.util.Map;
import java.util.HashMap;

/**
 *
 * @author S.Chassande-Barrioz
 */
public class TimeStamp extends Semaphore {

    /**
     * Time stamp of the object to which this object corresponds.
     */
    private long timeStamp;

    /**
     * Time stamp of the object to which this object corresponds, for each
     * context. This map associates to each context the timestamp value (in the
     * 63 higher order bits), and a flag indicating if the object has been
     * accessed in write mode by the context (in the lower order bit).
     */
    private Map ctxtTimeStamp;

    private int reservations = 0;

    public TimeStamp() {
        ctxtTimeStamp = new HashMap();
    }

    public void readIntention(Object ctxt) {
        if (ctxtTimeStamp.get(ctxt) == null) {
            ctxtTimeStamp.put(ctxt, new Long(timeStamp << 1));
        }
        reservations--;
    }

    public void writeIntention(Object ctxt) {
        Long lts = (Long) ctxtTimeStamp.get(ctxt);
        if (lts == null) {
            lts = new Long(timeStamp << 1);
            ctxtTimeStamp.put(ctxt, lts);
        }
        long l = lts.longValue();
        boolean dirty = (l & 1) != 0;
        if (!dirty) {
            long ts = l >>> 1;
            ctxtTimeStamp.put(ctxt, new Long((ts + 1) << 1 | 1));
        }
        reservations--;
    }

    public boolean validate(Object ctxt) {
        Long l = (Long) ctxtTimeStamp.get(ctxt);
        if (l != null) {
            long v = l.longValue();
            boolean dirty = (v & 1) != 0;
            long ts = (dirty ? (v >>> 1) - 1 : v >>> 1);
            if (ts != timeStamp) {
                return false;
            }
        }
        return true;
    }

    public void finalize(Object ctxt) {
        Long l = (Long) ctxtTimeStamp.get(ctxt);
        if (l != null) {
            timeStamp = l.longValue() >>> 1;
        }
    }

    /**
     * Removes the given context from the 'ctxtTimeStamp' map.
     * @param ctxt a context
     * @return true if the 'ctxtTimeStamp' map is empty, after the context has
     * been removed from the map. In such a case, this object can be removed
     * from the 'timeStamps' map of the enclosing class.
     */
    public boolean close(Object ctxt) {
        ctxtTimeStamp.remove(ctxt);
        return reservations == 0 && ctxtTimeStamp.isEmpty();
    }

    public void reserve() {
        reservations++;
    }
}
