/**
 * Copyright (C) 2001-2002
 * - France Telecom R&D
 * - Laboratoire Logiciels, Systemes, Reseaux - UMR 5526, CNRS-INPG-UJF
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Authors:
 *
 */

package org.objectweb.perseus.concurrency.pessimistic;

import org.objectweb.perseus.concurrency.api.ConcurrencyException;
import org.objectweb.perseus.concurrency.api.ConcurrencyManager;
import org.objectweb.perseus.dependency.api.DependencyGraph;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.util.monolog.api.Logger;

import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;

/**
 * @author E. Bruneton
 */
public abstract class PessimisticConcurrencyManager
        implements ConcurrencyManager, BindingController {

	public final static short POLICY_MUTEX = 1;
	public final static short POLICY_RW_PRIORITY2READERS = 2;
	public final static short POLICY_RW_FIFO = 3;

	public final static String DEPENDENCY_GRAPH_BINDING = "dependency-graph";

	/**
	 * True to distinguish read/write locks, false to use a single lock for
	 * both modes.
	 */
	protected short policy = POLICY_RW_FIFO;

	/**
	 * A map associating locks to all oids in all active contexts.
	 */
	protected Map locks;

    protected DependencyGraph dg = null;

	/**
	 * A map associating a ContextInfo to each context.
	 */
	protected Map contextInfos;

    protected Logger logger = null;

	public PessimisticConcurrencyManager() throws ConcurrencyException {
		this(POLICY_RW_FIFO);
	}

	public PessimisticConcurrencyManager(short policy) throws ConcurrencyException {
		super();
		locks = new HashMap();
		contextInfos = new HashMap();
		/*
		switch (policy) {
		case POLICY_MUTEX:
			logger.log(BasicLevel.INFO, "Pessimistic Concurrency Manager run with POLICY_MUTEX.");
			break;
		case POLICY_RW_PRIORITY2READERS:
			logger.log(BasicLevel.INFO, "Pessimistic Concurrency Manager run with POLICY_RW_PRIORITY2READERS.");
			break;
		case POLICY_RW_FIFO:
			logger.log(BasicLevel.INFO, "Pessimistic Concurrency Manager run with POLICY_RW_FIFO.");
			break;
		default:
			throw new ConcurrencyException("Unknown pessimistic concurrency policy.");
		}
		*/
		this.policy = policy;
	}

	/**
	 * @param ctx is the context using the resource
	 * @param resourceId is the resource identifier
	 * @param lock is the lock instance matching the resource
	 * @param hints uses to give information on access
	 * @return the resource state to use
	 */
	protected abstract Object getState(Object ctx,
									   Object resourceId,
									   Lock lock,
									   Object hints) throws ConcurrencyException;


    //IMPLEMENTATION OF THE UserBindingControler INTERFACE //
    //------------------------------------------------------//

    public String[] listFc() {
        return new String[] {DEPENDENCY_GRAPH_BINDING};
    }

    public Object lookupFc(String s) {
        if (DEPENDENCY_GRAPH_BINDING.equals(s)) {
            return dg;
        }
        return null;
    }

    public void bindFc(String s, Object o) {
        if ("logger".equals(s)) {
            logger = (Logger) o;
        } else if (DEPENDENCY_GRAPH_BINDING.equals(s)) {
			dg = (DependencyGraph) o;
        }
    }

    public void unbindFc(String s) {
		if (DEPENDENCY_GRAPH_BINDING.equals(s)) {
			dg = null;
		}
    }


	//IMPLEMENTATION OF THE ConcurrencyManager INTERFACE //
	//---------------------------------------------------//

	public void begin(Object ctx) {
		getContextInfo(ctx).rollback = false;
	}

	public boolean validate(Object ctx) {
		ContextInfo info = (ContextInfo) contextInfos.get(ctx);
		return info == null || !info.rollback;
	}

	public void finalize(Object ctx) {
		synchronized (locks) {
			Iterator entries = locks.entrySet().iterator();
			while (entries.hasNext()) {
				Map.Entry entry = (Map.Entry) entries.next();
				Lock lock = (Lock) entry.getValue();
				if (lock.close(ctx)) {
					entries.remove();
				}
			}
		}
		contextInfos.remove(ctx);
	}

	public void abort(Object ctx) {
		finalize(ctx);
	}

	public Object readIntention(Object ctx, Object resourceId, Object hints)
			throws ConcurrencyException {
		Lock lock = getLock(resourceId, hints);
        try {
            lock.readIntention(ctx);
        } catch (ConcurrencyException e) {
            getContextInfo(ctx).rollback = true;
            throw e;
        }
		return getState(ctx, resourceId, lock, hints);
	}

	public Object writeIntention(Object ctx, Object resourceId, Object hints)
			throws ConcurrencyException {
        Lock lock = getLock(resourceId, hints);
        try {
            lock.writeIntention(ctx);
        } catch (ConcurrencyException e) {
            getContextInfo(ctx).rollback = true;
            throw e;
        }
		return getState(ctx, resourceId, lock, hints);
	}

	//OTHER METHODS //
	//--------------//

	/**
	 * Returns the lock associated to the given oid.
	 * @param oid an object identifier.
	 * @return the lock associated to the given oid.
	 */
	protected Lock getLock(Object oid, Object hints) {
		synchronized (locks) {
			Lock lock = (Lock) locks.get(oid);
			if (lock == null) {
				switch (policy) {
				case POLICY_MUTEX:
					lock = new MutexLock(hints, dg);
					break;
				case POLICY_RW_PRIORITY2READERS:
					lock = new RWPri2RLock(hints, dg);
					break;
				case POLICY_RW_FIFO:
					lock = new RWFifoLock(hints, dg);
					break;
				}
				locks.put(oid, lock);
				//System.out.println("new ReadWriteLockValue(" + oid + "): " + lock);
			}
			lock.reserve();
			return lock;
		}
	}

	protected ContextInfo getContextInfo(Object ctxt) {
		ContextInfo info = (ContextInfo) contextInfos.get(ctxt);
		if (info == null) {
			info = new ContextInfo();
			contextInfos.put(ctxt, info);
		}
		return info;
	}

	/**
	 * Information associated to a context
	 */
	protected static class ContextInfo {
		// long date; // date of begin()
		public boolean rollback; // if the validate() method must return false
	}
}
