/**
 * Copyright (C) 2001-2002
 * - France Telecom R&D
 * - Laboratoire Logiciels, Systemes, Reseaux - UMR 5526, CNRS-INPG-UJF
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Authors:
 *
 */

package org.objectweb.perseus.concurrency.pessimistic;

import org.objectweb.perseus.concurrency.api.ConcurrencyException;
import org.objectweb.perseus.concurrency.api.RolledBackConcurrencyException;
import org.objectweb.perseus.concurrency.lib.ReadWriteLockValue;
import org.objectweb.perseus.dependency.api.DependencyGraph;

import java.util.Set;
import java.util.HashSet;
import java.util.List;
import java.util.ArrayList;

/**
 * A lock associated to an oid (see the "locks" map within the pessimistic
 * concurrency manager).
 * Provides "one writer/multiple readers" concurrency policy with a FIFO
 * scheduling.
 * @author E. Bruneton, P. Dechamboux
 */
public final class RWFifoLock extends Lock {
	/**
	 * The contexts that have a lock for the oid to which this object is
	 * associated.
	 */
	protected Object writer = null;
	/**
	 * The context that expect to upgrade from read to write lock.
	 */
	protected Object upgrader = null;
	/**
	 * The contexts that have a read lock for the oid to which this object
	 * is associated.
	 */
	protected Set readers = new HashSet();
	/**
	 * The contexts that wait for lock for the oid to which this object
	 * is associated.
	 */
	protected ArrayList waiters = new ArrayList();

	public RWFifoLock() {
	}

	public RWFifoLock(Object hints, DependencyGraph dg) {
		super(hints, dg);
	}

	/**
	 * Acquires this lock in read mode for the given context. This method
	 * blocks until the lock can be acquired in read mode by this context.
	 * @param ctxt a context.
	 */
	public synchronized void readIntention(Object ctxt)
			throws ConcurrencyException {
		boolean ok;
		ok = readers.contains(ctxt) || (writer == null
				? (waiters.size() == 0)
				: writer.equals(ctxt));
		if (!ok) {
			List l = new ArrayList();
			for (int i = 0; i < waiters.size(); i++) {
				l.add(((WaiterInfo) waiters.get(i)).getContext());
			}
			if (writer == null) {
				l.addAll(readers);
			} else {
				l.add(writer);
			}
			WaiterInfo wi = null;
			int pos = -1;
			try {
				for (int i = 0; i < l.size(); ++i) {
					if (!dg.addVertex(ctxt, l.get(i))) {
						throw new RolledBackConcurrencyException("Deadlock");
					}
				}
				wi = new WaiterInfo(true, ctxt);
				waiters.add(wi);
				do {
					//System.out.println(this + "@WaitRL: " + Thread.currentThread());
					wait();
					if (!waiters.contains(wi)) {
						wi = null;
						throw new RolledBackConcurrencyException(
								"The working set has been rolled back by another thread");
					}
					if (writer != null) {
						continue;
					}
					WaiterInfo cwi = null;
					for (pos = 0; pos < waiters.size(); pos++) {
						cwi = (WaiterInfo) waiters.get(pos);
						if (cwi == wi) {
							break;
						}
						if (!cwi.isReader()) {
							break;
						}
					}
					if (cwi == wi) {
						break;
					}
				} while (true);
			} catch (InterruptedException e) {
				throw new ConcurrencyException(
						"Waiting of a read intention has been interupted:", e);
			} finally {
				for (int i = 0; i < l.size(); ++i) {
					dg.removeVertex(ctxt, l.get(i));
				}
				if ((wi != null) && (pos != -1)) {
					waiters.remove(pos);
					//System.out.println(this + "@WakeupRL: " + Thread.currentThread());
				}
			}
		}
		reservations --;
		//System.out.println(this + "@GetRL: " + Thread.currentThread());
		readers.add(ctxt);
	}

	/**
	 * Acquires this lock in write mode for the given context. This method
	 * blocks until the lock can be acquired in write mode by this context.
	 * @param ctxt a context.
	 */
	public synchronized void writeIntention(Object ctxt)
			throws ConcurrencyException {
		boolean ok;
		if (writer == null) {
			int r = readers.size();
			ok = (waiters.size() == 0) && (r == 0 || (r == 1 && readers.contains(ctxt)));
		} else {
			ok = writer.equals(ctxt);
		}
		if (!ok) {
			if (readers.contains(ctxt)) {
				if (upgrader != null) {
					throw new RolledBackConcurrencyException("Deadlock on upgrade");
				}
				upgrader = ctxt;
			}
			List l = new ArrayList();
			int pos;
			for (pos = 0; pos < waiters.size(); pos++) {
				WaiterInfo cwi = (WaiterInfo) waiters.get(pos);
				if (!cwi.isReader()) {
					break;
				}
				l.add(cwi.getContext());
			}
			if (writer == null) {
				l.addAll(readers);
			} else {
				l.add(writer);
			}
			WaiterInfo wi = null;
			try {
				for (int i = 0; i < l.size(); ++i) {
					if (!dg.addVertex(ctxt, l.get(i))) {
						throw new RolledBackConcurrencyException("Deadlock");
					}
				}
				wi = new WaiterInfo(false, ctxt);
				waiters.add(pos, wi);
				do {
					//System.out.println(this + "@WaitWL: " + Thread.currentThread());
					wait();
					if (!waiters.contains(wi)) {
						wi = null;
						throw new RolledBackConcurrencyException(
								"The working set has been rolled back by another thread");
					}
					if (writer != null) {
						continue;
					}
					if ((WaiterInfo) waiters.get(0) == wi) {
						break;
					}
				} while (true);
			} catch (InterruptedException e) {
				throw new ConcurrencyException(
						"Waiting of a write intention has been interupted:", e);
			} finally {
				for (int i = 0; i < l.size(); ++i) {
					dg.removeVertex(ctxt, l.get(i));
				}
				if (wi != null) {
					waiters.remove(0);
					//System.out.println(this + "@WakeupWL: " + Thread.currentThread());
				}
			}
		}
		//System.out.println(this + "@GetWL: " + Thread.currentThread());
		if (upgrader == ctxt) {
			upgrader = null;
		}
		reservations --;
		writer = ctxt;
	}

	/**
	 * Removes the given context from the reader and writer lists of this
	 * lock.
	 * @param ctxt a context
	 * @return true if the reader and writer list are empty, after the
	 * context has been removed from these lists. In such a case, this
	 * object can be removed from the 'locks' map.
	 */
	public synchronized boolean close(Object ctxt) {
		readers.remove(ctxt);
		if (writer != null && writer.equals(ctxt)) {
			writer = null;
			//System.out.println(this + "@ReleaseWL: " + Thread.currentThread());
		} else {
			//System.out.println(this + "@ReleaseRL: " + Thread.currentThread());
		}
		for(int i=0; i<waiters.size();) {
			WaiterInfo wi = (WaiterInfo) waiters.get(i);
			if (wi.getContext() == ctxt) {
				waiters.remove(i);
			} else {
				i++;
			}			
		}
		int r = readers.size();
		if ((r == 0 || (r == 1 && upgrader != null)) && (waiters.size() != 0)) {
			notifyAll();
		}
		return reservations == 0 && waiters.isEmpty()
				&& readers.isEmpty() && writer == null;
	}


	public synchronized byte getMax() {
		if (writer != null) return ReadWriteLockValue.WRITE;
		if (!waiters.isEmpty()) {
			for (int cpt = 0; cpt < waiters.size(); cpt++)
				if (!((WaiterInfo) waiters.get(cpt)).isReader()) return ReadWriteLockValue.WRITE;
			return ReadWriteLockValue.READ;
		}
		if (!readers.isEmpty()) return ReadWriteLockValue.READ;
		return ReadWriteLockValue.NOLOCK;
	}

}

class WaiterInfo {
	private boolean reader;
	private Object context;

	WaiterInfo(boolean reader, Object context) {
		this.reader = reader;
		this.context = context;
	}

	boolean isReader() {
		return reader;
	}

	Object getContext() {
		return context;
	}
}
