/**
 * Copyright (C) 2001-2002
 * - France Telecom R&D
 * - Laboratoire Logiciels, Systemes, Reseaux - UMR 5526, CNRS-INPG-UJF
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Authors:
 *
 */

package org.objectweb.perseus.concurrency.pessimistic;

import org.objectweb.perseus.concurrency.api.ConcurrencyException;
import org.objectweb.perseus.concurrency.api.RolledBackConcurrencyException;
import org.objectweb.perseus.concurrency.lib.ReadWriteLockValue;
import org.objectweb.perseus.dependency.api.DependencyGraph;

import java.util.Set;
import java.util.HashSet;
import java.util.List;
import java.util.ArrayList;

/**
 * A lock associated to an oid (see the "locks" map within the pessimistic
 * concurrency manager).
 * Provides "one writer/multiple readers" concurrency policy with priority to
 * readers (there is a risk for starvation wrt writers).
 * @author E. Bruneton, P. Dechamboux
 */
public final class RWPri2RLock extends Lock {
	/**
	 * The contexts that have a lock for the oid to which this object is
	 * associated.
	 */
	protected Object writer = null;
	/**
	 * The number of threads that wait for aquiring this lock.
	 */
	int waiter = 0;
	/**
	 * The contexts that have a read lock for the oid to which this object
	 * is associated.
	 */
	protected Set readers = new HashSet();

	public RWPri2RLock() {
	}

	public RWPri2RLock(Object hints, DependencyGraph dg) {
		super(hints, dg);
	}

	/**
	 * Acquires this lock in read mode for the given context. This method
	 * blocks until the lock can be acquired in read mode by this context.
	 * @param ctxt a context.
	 */
	public synchronized void readIntention(Object ctxt)
			throws ConcurrencyException {
		boolean ok;
		do {
			ok = writer == null || writer.equals(ctxt);
			if (!ok) {
				waiter ++;
				Object w = writer;
				try {
					if (!dg.addVertex(ctxt, w)) {
						throw new RolledBackConcurrencyException("Deadlock");
					}
					wait();
				} catch (InterruptedException e) {
					throw new ConcurrencyException(
						"Waiting of a read intention has been interupted:", e);
				} finally {
					waiter --;
					dg.removeVertex(ctxt, w);
				}
			}
		} while (!ok);
		reservations--;
		readers.add(ctxt);
//TODO implements writeIntention
	}

	/**
	 * Acquires this lock in write mode for the given context. This method
	 * blocks until the lock can be acquired in write mode by this context.
	 * @param ctxt a context.
	 */
	public synchronized void writeIntention(Object ctxt)
			throws ConcurrencyException {
		boolean ok;
		do {
			if (writer == null) {
				int r = readers.size();
				ok = r == 0 || (r == 1 && readers.contains(ctxt));
			} else {
				ok = writer.equals(ctxt);
			}
			if (!ok) {
				List l = new ArrayList();
				if (writer == null) {
					l.addAll(readers);
				} else {
					l.add(writer);
				}
				waiter ++;
				try {
					for (int i = 0; i < l.size(); ++i) {
						if (!dg.addVertex(ctxt, l.get(i))) {
							throw new RolledBackConcurrencyException("Deadlock");
						}
					}
					wait();
				} catch (InterruptedException e) {
					throw new ConcurrencyException(
						"Waiting of a write intention has been interupted:", e);
				} finally {
					waiter --;
					for (int i = 0; i < l.size(); ++i) {
						dg.removeVertex(ctxt, l.get(i));
					}
				}
			}
		} while (!ok);
		reservations--;
		writer = ctxt;
//TODO implements writeIntention
	}

	/**
	 * Removes the given context from the reader and writer lists of this
	 * lock.
	 * @param ctxt a context
	 * @return true if the reader and writer list are empty, after the
	 * context has been removed from these lists. In such a case, this
	 * object can be removed from the 'locks' map.
	 */
	public synchronized boolean close(Object ctxt) {
		//TODO: wake up blocking threads working in the closed context
		readers.remove(ctxt);
		if (writer != null && writer.equals(ctxt)) {
			writer = null;
		}
		boolean res = reservations == 0
				&& waiter == 0 && readers.isEmpty() && writer == null;
		if (!res) {
			notifyAll();
		}
		return res;
	}

	public synchronized byte getMax() {
		if (writer != null) return ReadWriteLockValue.WRITE;
		if (!readers.isEmpty()) return ReadWriteLockValue.READ;
		return ReadWriteLockValue.NOLOCK;
	}

}
